;;;; spread.scm

#>
#include <sp.h>
<#

#>?
const int MAX_PRIVATE_NAME, MAX_GROUP_NAME;

const int ACCEPT_SESSION;
const int ILLEGAL_SPREAD;
const int COULD_NOT_CONNECT;
const int REJECT_QUOTA;
const int REJECT_NO_NAME;
const int REJECT_ILLEGAL_NAME;
const int REJECT_NOT_UNIQUE;
const int REJECT_VERSION;
const int CONNECTION_CLOSED;
const int REJECT_AUTH;

const int ILLEGAL_SESSION;
const int ILLEGAL_SERVICE;
const int ILLEGAL_MESSAGE;
const int ILLEGAL_GROUP;
const int BUFFER_TOO_SHORT;
const int GROUPS_TOO_SHORT;
const int MESSAGE_TOO_LONG;

const int DEFAULT_SPREAD_PORT;

const int
  REG_MEMB_MESS,
  TRANSITION_MESS,
  CAUSED_BY_JOIN,
  CAUSED_BY_LEAVE,
  CAUSED_BY_DISCONNECT,
  CAUSED_BY_NETWORK,
  MEMBERSHIP_MESS;

const int 
  SELF_DISCARD,
  DROP_RECV,

  UNRELIABLE_MESS,
  RELIABLE_MESS,
  FIFO_MESS,
  CAUSAL_MESS,
  AGREED_MESS,
  SAFE_MESS,
  REGULAR_MESS;

const int
  LOW_PRIORITY, MEDIUM_PRIORITY, HIGH_PRIORITY;

int SP_connect(char *, char *, int, ___bool, void *, void *);
int SP_disconnect(int);
int SP_join(int, char *);
int SP_leave(int, char *);
___bool SP_poll(int);
int SP_multicast(int, int, char *, int, int, char *);
int SP_multigroup_multicast(int, int, int, void *, int, int, char *);
int SP_receive(int, void *, void *, int, void *, void *, void *, void *, int, void *);
int SP_version(void *, void *, void *);
<#

#>!
static char *strunc(___scheme_value s) { return C_c_string(s); }
<#

(declare
  (uses srfi-1)
  (export sp:connection-private-group sp:connection? sp:version
	  sp:connect sp:disconnect sp:join sp:leave sp:poll sp:multicast sp:receive
	  sp:message? sp:message-data sp:message-endian-mismatch?
	  sp:regular-message? sp:regular-membership-message? sp:transitional-membership-message? sp:self-leave-message?
	  sp:message-sender sp:message-groups sp:message-service-type sp:message-type) )

(define-record sp:connection
  mailbox				; int
  private-group)			; string

(define (sp:connect #!optional (spread-name (number->string DEFAULT_SPREAD_PORT))
		    #!key (private-name (random-name)) (priority MEDIUM_PRIORITY) (group-membership #t) )
  (let ([pgroup (make-string (add1 MAX_GROUP_NAME))])
    (let-location ([mbox int])
      (check (SP_connect spread-name private-name priority group-membership #$mbox #$pgroup) 'sp:connect)
      (make-sp:connection mbox (strunc pgroup)) ) ) )

(define (random-name)
  (apply string (list-tabulate MAX_PRIVATE_NAME (lambda _ (integer->char (+ (char->integer #\0) (random 64)))))) )

(define error-codes
  `((,ACCEPT_SESSION "accept session")
    (,ILLEGAL_SPREAD "illegal spread")
    (,COULD_NOT_CONNECT "could not connect")
    (,REJECT_QUOTA "reject quota")
    (,REJECT_NO_NAME "reject no name")
    (,REJECT_ILLEGAL_NAME "reject illegal name")
    (,REJECT_NOT_UNIQUE "reject not unique")
    (,REJECT_VERSION "reject version")
    (,CONNECTION_CLOSED "connection closed")
    (,REJECT_AUTH "reject auth")
    (,ILLEGAL_SESSION "illegal session")
    (,ILLEGAL_SERVICE "illegal service")
    (,ILLEGAL_MESSAGE "illegal message")
    (,ILLEGAL_GROUP "illegal group")
    (,BUFFER_TOO_SHORT "buffer too short")
    (,GROUPS_TOO_SHORT "groups too short")
    (,MESSAGE_TOO_LONG "message too long") ) )

(define (check result loc)
  (when (< result 0)
    (let ([msg (assq result error-codes)])
      (signal
       (make-composite-condition
	(make-property-condition 'exn 'message (if msg (cadr msg) "???") 'location loc 'arguments '())
	(make-property-condition 'spread 'code result) ) ) ) )
  result)

(define (sp:disconnect con)
  (check (SP_disconnect (sp:connection-mailbox con)) 'sp:disconnect) )

(define (sp:join con group)
  (check (SP_join (sp:connection-mailbox con) group) 'sp:join) )

(define (sp:leave con group)
  (check (SP_leave (sp:connection-mailbox con) group) 'sp:leave) )

(define (sp:poll con)
  (SP_poll (sp:connection-mailbox con)) )

(define service-types
  `((unreliable ,UNRELIABLE_MESS)
    (reliable ,RELIABLE_MESS)
    (fifo ,FIFO_MESS)
    (causal ,CAUSAL_MESS)
    (agreed ,AGREED_MESS)
    (safe ,SAFE_MESS)
    (regular ,REGULAR_MESS)
    (regular-membership ,REG_MEMB_MESS)
    (transition ,TRANSITION_MESS)
    (caused-by-join ,CAUSED_BY_JOIN)
    (caused-by-leave ,CAUSED_BY_LEAVE)
    (caused-by-disconnect ,CAUSED_BY_DISCONNECT)
    (caused-by-network ,CAUSED_BY_NETWORK)
    (membership ,MEMBERSHIP_MESS) ) )

#>!
static void *make_group_buffer(___scheme_value groups, int n)
{
  char *buf = (char *)malloc(n * (MAX_GROUP_NAME)), *pbuf = buf;
  int i;
  ___scheme_value g;

  for(i = 0; i < n; ++i) {
    if(groups != C_SCHEME_END_OF_LIST) {
      memset(buf, 0, MAX_GROUP_NAME);
      g = C_block_item(groups, 0);
      memcpy(buf, C_c_string(g), C_header_size(g));
      groups = C_block_item(groups, 1);
    }

    buf += MAX_GROUP_NAME;
  }

  return pbuf;
}


static void free_group_buffer(void *b)
{
  free(b);
}
<#

(define (sp:multicast con service-type group msg #!key (self-discard #t) (message-type 0))
  (let ([mbox (sp:connection-mailbox con)]
	[stype (assq service-type service-types)] 
	[len (string-length msg)] )
    (if stype
	(set! stype (cadr stype))
	(error 'sp:multicast "invalid service type" service-type) )
    (when self-discard (set! stype (bitwise-ior SELF_DISCARD stype)))
    (check
     (if (list? group)
	 (let* ([n (length group)]
		[gbuffer (make_group_buffer group n)] 
		[r (SP_multigroup_multicast mbox stype n gbuffer message-type len msg)] )
	   (free_group_buffer gbuffer)
	   r)
	 (SP_multicast mbox stype group message-type len msg) ) 
     'sp:multicast) ) )

(define-constant default-max-groups 16)
(define-constant default-message-length 1024)

(define-record sp:message
  service-type				; (symbol ...)
  sender				; string
  groups				; (string ...)
  type					; int
  endian-mismatch?			; bool
  data)					; string | (id member ...)

(define-external (cons_string (c-string name) (scheme-object lst)) scheme-object
  (cons name lst) )

(define-external (cons_number (integer n) (scheme-object lst)) scheme-object
  (cons n lst) )

#>!
___scheme_value cons_string(char *, ___scheme_value);
___scheme_value cons_number(int, ___scheme_value);

static ___safe ___scheme_value gather_groups(void *gbuffer, int n)
{
  char *gb = (char *)gbuffer;
  ___scheme_value lst = C_SCHEME_END_OF_LIST;

  while(n--) {
    lst = cons_string(gb, lst);
    gb += MAX_GROUP_NAME;
  }

  return lst;  
}

static ___safe ___scheme_value gather_membership_info(void *mbuffer)
{
  struct { 
    group_id id;
    int num_groups;
    char members[ 1 ][ MAX_GROUP_NAME ];
  } *minfo = mbuffer;
  ___scheme_value lst = C_SCHEME_END_OF_LIST;
  int n = minfo->num_groups, i;

  for(i = 0; i < n; ++i)
    lst = cons_string(minfo->members[ i ], lst);

  for(n = 0; n < 3; ++n) 
    lst = cons_number(minfo->id.id[ i ], lst);
  
  return lst;
}
<#

(define (sp:receive con #!key
		    (max-groups default-max-groups)
		    (drop-recv #f) 
		    max-message-length
		    (destination (make-string (or max-message-length default-message-length))) )
  (let-location ([stype int (if drop-recv DROP_RECV 0)]
		 [ngroups int]
		 [mtype short]
		 [emismatch int] )
    (let ([sender (make-string MAX_GROUP_NAME)]
	  [gbuffer (make_group_buffer '() max-groups)] )
      (let loop ()
	(let ([size
	       (SP_receive
		(sp:connection-mailbox con)
		#$stype #$sender max-groups #$ngroups gbuffer #$mtype #$emismatch 
		(or max-message-length (string-length destination))
		#$destination) ] )
	  (switch size
	    [GROUPS_TOO_SHORT 
	     (free_group_buffer gbuffer)
	     (set! max-groups (* 2 max-groups))
	     (cond-expand [debug (print "[sp:receive - resizing groups to " max-groups "]")][else])
	     (loop) ]
	    [BUFFER_TOO_SHORT
	     (if max-message-length
		 (check size 'sp:receive)
		 (begin
		   (set! destination (make-string (* 2 (string-length destination))))
		   (cond-expand [debug (print "[sp:receive - resizing destination to " (string-length destination) "]")]
				[else])
		   (loop) ) ) ]
	    [else
	     (if (< size 0)
		 (check size 'sp:receive)
		 (let ([types 
			(let loop ([sts service-types])
			  (cond [(null? sts) '()]
				[(not (zero? (bitwise-and (cadar sts) stype))) 
				 (cons (caar sts) (loop (cdr sts))) ]
				[else (loop (cdr sts))] ) ) ] )
		   (make-sp:message 
		    types
		    (strunc sender)
		    (gather_groups gbuffer ngroups)
		    mtype
		    (if (zero? emismatch) #f #t)
		    (if (zero? (bitwise-and stype REG_MEMB_MESS))
			(substring destination 0 size)
			(gather_membership_info #$destination) ) ) ) ) ] ) ) ) ) ) )

(define (sp:regular-message? m)
  (memq 'regular (sp:message-service-type m)) )

(define (sp:regular-membership-message? m)
  (memq 'regular-membership (sp:message-service-type m)) )

(define (sp:transitional-membership-message? m)
  (memq 'transition (sp:message-service-type m)) )

(define (sp:self-leave-message? m)
  (let ([sts (sp:message-service-type m)])
    (and (memq 'caused-by-leave sts)
	 (not (memq 'regular-membership sts)) ) ) )

(define (sp:version)
  (let-location ([maj int] [min int] [p int])
    (SP_version #$maj #$min #$p)
    (values maj min p) ) )
